@@ -935,6 +935,12 @@ int BrowserService::sortPriority(const Entry* entry,
     if (url.scheme().isEmpty()) {
         url.setScheme("https");
     }
+
+    // Add the empty path to the URL if it's missing
+    if (url.path().isEmpty() && !url.hasFragment() && !url.hasQuery()) {
+        url.setPath("/");
+    }
+
     const QString entryURL = url.toString(QUrl::StripTrailingSlash);
     const QString baseEntryURL =
         url.toString(QUrl::StripTrailingSlash | QUrl::RemovePath | QUrl::RemoveQuery | QUrl::RemoveFragment);
@@ -1046,7 +1052,12 @@ bool BrowserService::handleURL(const QString& entryUrl, const QString& url, cons
         return false;
     }
 
-    // Filter to match hostname in URL field
+    // Match the base domain
+    if (baseDomain(siteQUrl.host()) != baseDomain(entryQUrl.host())) {
+        return false;
+    }
+
+    // Match the subdomains with the limited wildcard
     if (siteQUrl.host().endsWith(entryQUrl.host())) {
         return true;
     }
