@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2018 KeePassXC Team <team@keepassxc.org>
+ *  Copyright (C) 2019 KeePassXC Team <team@keepassxc.org>
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -21,11 +21,14 @@
 #include "core/Database.h"
 #include "util/TemporaryFile.h"
 
+#include <QByteArray>
 #include <QFile>
 #include <QScopedPointer>
+#include <QSharedPointer>
 #include <QTemporaryFile>
 #include <QTest>
-#include <QTextStream>
+
+#include <stdio.h>
 
 class TestCli : public QObject
 {
@@ -40,41 +43,55 @@ private slots:
     void cleanup();
     void cleanupTestCase();
 
-    void testCommand();
+    void testBatchCommands();
     void testAdd();
+    void testAddGroup();
+    void testAnalyze();
     void testClip();
+    void testCommandParsing_data();
+    void testCommandParsing();
     void testCreate();
     void testDiceware();
     void testEdit();
     void testEstimate_data();
     void testEstimate();
-    void testExtract();
+    void testExport();
     void testGenerate_data();
     void testGenerate();
+    void testImport();
     void testKeyFileOption();
     void testNoPasswordOption();
+    void testHelp();
+    void testInteractiveCommands();
     void testList();
     void testLocate();
     void testMerge();
+    void testMergeWithKeys();
+    void testMove();
+    void testOpen();
     void testRemove();
+    void testRemoveGroup();
     void testRemoveQuiet();
     void testShow();
+    void testInvalidDbFiles();
+    void testYubiKeyOption();
 
 private:
     QByteArray m_dbData;
     QByteArray m_dbData2;
+    QByteArray m_xmlData;
+    QByteArray m_yubiKeyProtectedDbData;
     QByteArray m_keyFileProtectedDbData;
     QByteArray m_keyFileProtectedNoPasswordDbData;
     QScopedPointer<TemporaryFile> m_dbFile;
     QScopedPointer<TemporaryFile> m_dbFile2;
+    QScopedPointer<TemporaryFile> m_xmlFile;
     QScopedPointer<TemporaryFile> m_keyFileProtectedDbFile;
     QScopedPointer<TemporaryFile> m_keyFileProtectedNoPasswordDbFile;
+    QScopedPointer<TemporaryFile> m_yubiKeyProtectedDbFile;
     QScopedPointer<TemporaryFile> m_stdoutFile;
     QScopedPointer<TemporaryFile> m_stderrFile;
     QScopedPointer<TemporaryFile> m_stdinFile;
-    FILE* m_stdoutHandle = stdout;
-    FILE* m_stderrHandle = stderr;
-    FILE* m_stdinHandle = stdin;
 };
 
 #endif // KEEPASSXC_TESTCLI_H
